package ca.umontreal.iro.rali;

import java.util.HashMap;
import java.util.Map;

/**
 * Parsing of command line options and keep their values that can be shared by the Explore and Extract classes.
 * 
 * @author Guy Lapalme, Université de Montréal, 2009
 *
 */
public class Options {
    protected static final String XML_NS = "http://www.w3.org/2000/xmlns/";
	
	// values set by command line arguments
    // common
	/**
	 * Global option (-d): will show debugging information during execution
	 */
	public static boolean debug=false;
	/**
	 * Global option (-st): will show statistics information at the end of execution 
	 */
	protected static boolean statistics=false;
	/**
	 * Last parameter: Name of the input file, if it is null then standard input is used
	 */
	protected static String inputFileName=null;
	/**
	 * Global option (-o): Name of the output file, if it is null then standard output is used
	 */
	protected static String outputFileName=null;
	
	protected static boolean sax=false;
	
	// Xplore
	/**
	 * Explore option (-a): output attribute names in XPath expression
	 */
	protected static boolean outAttr=false;
	/**
	 * Explore option (-av): output attribute names and values in XPath expression
	 */
	protected static boolean outAttrVal=false;
	/**
	 * Explore option (-v): output element values in XPath expression
	 */
	protected static boolean outVal=false;
	/**
	 * Explore option (-sa): sort the XPath expressions alphabetically
	 */
	protected static boolean sortAlpha=false;
	/**
	 * Explore option (-sf): sort the XPath expression by decreasing frequency
	 */
	protected static boolean sortFreq=false;
	/**
	 * Explore option (-ns): show the values of the namespaces used in XPath expressions
	 */
	protected static boolean showNamespaces=false;
	// Xtract
	/**
	 * Extract option (-xp): XPath expression used to select nodes to extract
	 */
	protected static String xpathStr=null;
	/**
	 * Extract option (-xsl): XSLT stylesheet to apply to each extracted node
	 */
	protected static String xsltFile=null;
	/**
	 * Extract option (-xslf): XSLT stylesheet to apply to the resulting document
	 */
	protected static String xsltFileFinal=null;
	
	/**
	 * List of parameters for the xsl stylesheets (they are set to both individual and final stylesheets) 
	 */
	protected static Map<String,String> xslParams;
	
	/**
	 * Extract option (-r): name of the root element (<code>root</code> if not given)
	 */
	protected static String rootName="root";
	/**
	 * Extract option (-c): return only the number of extracted nodes
	 */
	protected static boolean onlyCount=false;
	/**
	 * Extract option (-pp): output a prettyprinted version of the extracted node
	 */
	protected static boolean prettyPrint=false;
	/**
	 * Extract option (-max): stop reading the file after a given number of results (infinite if not specified)
	 */
	protected static int maxRes = -1; // ie infinite...
	
	
	/**
	 * display of usage of the program with the list of possible command line arguments
	 */
	private static String usageStr="usage: [-h] [-d] [-st] [-o output file name]" + // common options
                                          "[-a] [-av] [-v] [-sf] [-sa] [-ns] " + // xplore options
                                          // xtract options
                                          "[-xp XPathExpr] [-xpt XPathExpr] [-max N] [-xsl xsltFile] [-xslf xsltFile] [-r rootname] [-nr] [-c] [-pp] {-xmlns:pref=URI} "+
                                          "[input file name] "+
                                          "{param_i=val_i}";
	private static String moreHelp="\n Use -h option for more details";

	/**
	 * Parse program arguments and set appropriate options
	 * will stop on -h : prints usage of the program 
	 *              -xpt: prints the xpath expression as it is seen by the program
	 * will stop on an unrecognized arguments      
	 * 
	 * @param args String array as given to the main program
	 */
	protected static void parseArgs(String[] args){
		int i=0;
		String arg;
		while (i<args.length && args[i].startsWith("-")){
			arg=args[i++];
			// common options
			if (arg.equals("-h")) usage();
			else if (arg.equals("-d")) debug=true;
			else if (arg.equals("-st")) statistics=true;
			else if (arg.equals("-sax")) sax=true;
			else if (arg.equals("-o")) outputFileName=args[i++];
			// xplore options
			else if (arg.equals("-a")) outAttr=true;
			else if (arg.equals("-av"))outAttrVal=true;
			else if (arg.equals("-v")) outVal=true;
			else if (arg.equals("-sa"))sortAlpha=true;
			else if (arg.equals("-sf"))sortFreq=true;
			else if (arg.equals("-ns"))showNamespaces=true;
			// xtract options
			else if (arg.equals("-xp"))xpathStr=args[i++];
			else if (arg.equals("-xpt")){
				System.out.println("XPathExpr as seen by the program>"+args[i++]+"<");
				System.exit(1);
			}
			else if (arg.equals("-xsl"))xsltFile=args[i++];
			else if (arg.equals("-xslf"))xsltFileFinal=args[i++];
			else if (arg.equals("-r")) rootName=args[i++];
			else if (arg.equals("-nr"))rootName=null;
			else if (arg.equals("-c")) onlyCount=true;
			else if (arg.equals("-pp")) prettyPrint=true;
			else if (arg.equals("-max")){
				try {
					maxRes=Integer.parseInt(args[i++]);
				} catch (NumberFormatException e){
					System.out.println("Error in number parsing for -max option :"+args[i-1]);
				}
			} else if (arg.startsWith("-xmlns:")){
				String[]nsparts=arg.substring(7).split("=");
				if(nsparts.length!=2){
					System.err.println("-xmlns option should of the form xmlns:prefix=URI");
					System.exit(1);
				}
				nsContext.addPrefixURI(nsparts[0], nsparts[1]);
			}
			else {
				System.err.println("illegal option:"+arg);
				System.err.println(usageStr+moreHelp);
				System.exit(1);
			}
		}
		if(i<args.length){
			inputFileName=args[i++];
			xslParams=new HashMap<String,String>();
			while (i<args.length){
				String[] s = (args[i++]).split("=");
				if(s.length==2)
					xslParams.put(s[0],s[1]);
				else
					System.err.print("Illegal parameter:"+args[i-1]);
			}
		}
	}

	/**
	 * Prints a short help about the use of the program
	 * called with the -h program argument
	 * 
	 */
	private static void usage(){
		System.out.println("Xplore or Xtract XML files");
		System.out.println(usageStr);
		System.out.println("Common options");
		System.out.println(" -h: print this help on standard output and exit");
		System.out.println(" -d: print debugging information on the standard output");
		System.out.println("-sax: use sax parser instead of pull");
		System.out.println("-st: print some program execution statistics on the standard error");
		System.out.println(" -o: name of the output file for the results, standard output if none is specified");
		System.out.println("Xplore options (i.e. when no -xp is specified)");
		System.out.println(" -a: show attributes for elements");
		System.out.println("-av: show attribute values for elements (implies -a)");
		System.out.println(" -v: show values of text nodes");
		System.out.println("-sa: sort xpath expressions alphabetically");
		System.out.println("-sf: sort xpath expressions by decreasing freauency");
		System.out.println("-ns: display namespaces");
		System.out.println("Xtract options (when -xp or xpt is specified) ");
		System.out.println("-xp: XPath expression of nodes to output, whitespace nodes are ignored");
		System.out.println("     the expression must return a NODESET and should not make 'outside' references");
		System.out.println("-xpt: print the XPath expression and exit; useful to see the expression as seen by the program");
		System.out.println("-xsl: name of XSLT transformation file with templates to apply to each result node");
		System.out.println("-xslf: name of a final XSLT transformation file with templates to apply to the overall document");
		System.out.println(" -r: name of root element, 'root' if not specified");
		System.out.println("-nr: do not output root node");
		System.out.println(" -c: output only count of output nodes");
		System.out.println("-pp: pretty print the XML result nodes");
		System.out.println("-max: next arg is a positive integer giving the maximum number of results needed");
		System.out.println("-xmlns:pref=URI : prefix and corresponding namespace to add to the root (can be repeated)");
		System.out.println("fileName : name of input file, standard input if none is specified");
		System.out.println("params=vals: list of parameter names and their value to set for the stylesheets");
		System.exit(1);
	}

	/**
	 * keep track of the namespaces used in the input document
	 */
	protected static NamespaceContext nsContext = new NamespaceContext();

}
